/************************************************************************/
/*                                                                      */
/*                                                                      */
/*    *****                       *****                                 */
/*      *****                   *****                                   */
/*        *****               *****                                     */
/*          *****           *****                                       */
/*  ***************       ***************                               */
/*  *****************   *****************                               */
/*  ***************       ***************                               */
/*          *****           *****           TheNetNode                  */
/*        *****               *****         Portable                    */
/*      *****                   *****       Network                     */
/*    *****                       *****     Software                    */
/*                                                                      */
/* This file is part of "TheNetNode" - Software Package                 */
/*                                                                      */
/* Copyright (C) 1998  NORD><LINK e.V. Braunschweig                     */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the NORD><LINK ALAS (Allgemeine Lizenz fr     */
/* Amateurfunk Software) as published by Hans Georg Giese (DF2AU)       */
/* on 13/Oct/1992; either version 1, or (at your option) any later      */
/* version.                                                             */
/*                                                                      */
/* This program is distributed WITHOUT ANY WARRANTY only for further    */
/* development and learning purposes. See the ALAS (Allgemeine Lizenz   */
/* fr Amateurfunk Software).                                           */
/*                                                                      */
/* You should have received a copy of the NORD><LINK ALAS (Allgemeine   */
/* Lizenz fr Amateurfunk Software) along with this program; if not,    */
/* write to NORD><LINK e.V., Hinter dem Berge 5, D-3300 Braunschweig    */
/*                                                                      */
/* Dieses Programm ist PUBLIC DOMAIN, mit den Einschrnkungen durch     */
/* die ALAS (Allgemeine Lizenz fr Amateurfunk Software), entweder      */
/* Version 1, verffentlicht von Hans Georg Giese (DF2AU),              */
/* am 13.Oct.1992, oder (wenn gewnscht) jede sptere Version.          */
/*                                                                      */
/* Dieses Programm wird unter Haftungsausschlu vertrieben, aus-        */
/* schlielich fr Weiterentwicklungs- und Lehrzwecke. Nheres          */
/* knnen sie der ALAS (Allgemeine Lizenz fr Amateurfunk Software)     */
/* entnehmen.                                                           */
/*                                                                      */
/* Sollte dieser Software keine ALAS (Allgemeine Lizenz fr Amateurfunk */
/* Software) beigelegen haben, wenden Sie sich bitte an                 */
/* NORD><LINK e.V., Hinter dem Berge 5, D-38108 Braunschweig            */
/*                                                                      */
/*                                                                      */
/************************************************************************/

#include "tnn.h"

/* Layer 4 parameter ---------------------------------------------------*/

#define L4_ACKDEL    2          /* L4 Acknowledge Wartezeit fr Circuit */
#define L4_BSYDEL  180          /* L4 Busy-Wartezeit in Sekunden        */

UWORD   l4_beta1 = 3;           /* RETRANS-TIMER (T1) = SRTT * BETA1    */
UWORD   l4_beta2 = 1;           /* ACK-TIMER (T2) = SRTT * BETA2        */
UWORD   l4_beta3 = 20;          /* BUSY/REQ-TIMEOUT (T3) = SRTT * BETA3 */
#define L4_BETA1 (l4_beta1)
#define L4_BETA2 (l4_beta2)
#define L4_BETA3 (l4_beta3)

#define L4_ALPHA1    2          /* Faktor fuer steigenden L4RTT         */
#define L4_ALPHA2    3          /* Faktor fuer fallenden L4RTT          */
#define L4_IRTT  1000U          /* Startwert SRTT (IRTT)                */

#define L4_RETRY     3          /* Anzahl der maximalen Retries         */

/*----------------------------------------------------------------------*/

LHEAD l4frel;                   /* freie L4-Controllbloecke             */
LHEAD l4actl;                   /* benutzte L4-Controllebloecke         */

/*----------------------------------------------------------------------*/
void initl4(void)                     /* Level 4 initialisieren         */
{
  int i;                              /* Sratch Zaehler                 */

  inithd(&l4rxfl);                    /* Liste fuer empfangene Frames   */
  inithd(&l4frel);                    /* freie L4-Controllbloecke       */
  inithd(&l4actl);                    /* benutzte L4-Controllebloecke   */

  for (i = 0, cirpoi = cirtab;        /* gesamte Circuit Tabelle        */
       i < NUMCIR;                    /* bearbeiten                     */
       ++i, ++cirpoi)
  {
    cirpoi->state  = L4SDSCED;        /* Eintrag ist leer               */
    cirpoi->numrx  = 0;               /* keine Frames empfangen         */
    cirpoi->numtx  = 0;               /* keine Frames zu senden         */
    cirpoi->ll4txNR  = 0;             /* last l4 ack sent               */
    cirpoi->fragme = NULL;            /* kein Frame Fragment da         */
    inithd(&cirpoi->mbhdrx);          /* Empfangskette ist leer         */
    inithd(&cirpoi->mbhdtx);          /* Sendekette ist leer            */
    inithd(&cirpoi->mbhdos);          /* Kette "ausser Reihenfolge"     */
    clrcir();
    relink((LEHEAD *)cirpoi,          /* in die Liste fuer unbenutzte   */
           (LEHEAD *)l4frel.tail);    /* Controllbloecke                */
  }
  nmbcir = nmbcir_max = 0;            /* Circuit-Zaehler                */
}

/*----------------------------------------------------------------------*/
#define START 0
#define STOP  1
#define CLEAR 2
void l4rtt(int what)
{
  switch (what) {
    case START : cirpoi->RTT = 1;
                 cirpoi->RTTvs = cirpoi->l4vs;
                 break;

    case STOP  : if (cirpoi->RTT > cirpoi->SRTT)
                   cirpoi->SRTT = (L4_ALPHA1 * cirpoi->SRTT +
                                   cirpoi->RTT * 100) / (L4_ALPHA1 + 1);
                 else
                   cirpoi->SRTT = (L4_ALPHA2 * cirpoi->SRTT +
                                   cirpoi->RTT * 100) / (L4_ALPHA2 + 1);
                 if (cirpoi->SRTT < L4_IRTT / 10)
                   cirpoi->SRTT = L4_IRTT / 10;
                 if (cirpoi->SRTT > L4_IRTT * 10)
                   cirpoi->SRTT = L4_IRTT * 10;

    case CLEAR : cirpoi->RTT =
                 cirpoi->RTTvs = 0;
                 break;
  }
}

/************************************************************************/
/*                                                                      */
/* RETRY TIMER (T1)                                                     */
/*                                                                      */
/* Der T1 wird fr jedes Frame in der Sendewarteschlange verwaltet und  */
/* bei jeder Aussendung dieses Frames gestartet. Wenn er abluft, wird  */
/* das Frame erneut ausgesendet. L4_BETA1 mu so gewhlt werden, da    */
/* der T1 nicht zu frh abluft und damit das Frame doppelt geschickt   */
/* wird.                                                                */
/*                                                                      */
/************************************************************************/
void l4setT1(MBHEAD *fbp) {
  if (fbp->l4trie > 1)
    fbp->l4time = (cirpoi->SRTT/100 + 1) * 2 * L4_BETA1;
  else
    fbp->l4time = (cirpoi->SRTT/100 + 1) * L4_BETA1;
  if (fbp->l4time < 30)             /* nicht unter 30s                  */
    fbp->l4time = 30;
}

/************************************************************************/
/*                                                                      */
/* ACKNOWLEDGE TIMER (T2)                                               */
/*                                                                      */
/* Der T2 bestimmt die maximale Verzgerung, bis ein empfangenes Frame  */
/* besttigt wird. L4_BETA sollte so gewhlt werden, das Frames         */
/* mglichst schnell besttigt werden, aber nicht zuviele ACK-Frames    */
/* generiert werden.                                                    */
/*                                                                      */
/************************************************************************/
void l4setT2(void) {
  UWORD frack;

  frack = (cirpoi->SRTT/100 + 1) * L4_BETA2;

  if (frack < 10)
    frack = 10;

  if (   cirpoi->acktim > frack
      && cirpoi->numrx  < (cirpoi->window /2)
     ) cirpoi->acktim = frack;
}

/************************************************************************/
/*                                                                      */
/* CONREQ/BUSY TIMEOUT (T3)                                             */
/*                                                                      */
/* Der T3 bestimmt die Wartezeit zwischen der Wiederholung von CONACK/  */
/* DISREQ und der Linkkontrolle bei Busy.                               */
/* Der T3 sollte nicht zu gering gewhlt werden.                        */
/*                                                                      */
/************************************************************************/
void l4setT3(int what) {
#define L4TCREQ  0
#define L4TDREQ  1
#define L4TBUSY  2
  switch (what) {
    case L4TCREQ  :
    case L4TDREQ  : cirpoi->traout = (cirpoi->SRTT/100 + 1) * L4_BETA3;
                    if (cirpoi->traout < 30)
                        cirpoi->traout = 30;
                    break;
    case L4TBUSY  : cirpoi->traout = L4_BSYDEL;
                    break;
  }
}

/*----------------------------------------------------------------------*/
void l4clrT3(void) {
  cirpoi->traout = 0;
}

/*----------------------------------------------------------------------*/
void l4newstate(UWORD state)
{
  if (state) {                         /* Verbindung steht              */
    if (cirpoi->state == L4SDSCED) {   /* ein neuer Connect             */
      ulink((LEHEAD *)cirpoi);         /* Circuit aus der Freiliste     */
      relink((LEHEAD *)cirpoi,         /* in die Liste fuer benutzte    */
             (LEHEAD *)l4actl.tail);   /* Controllbloecke               */
      if (++nmbcir > nmbcir_max)       /* Maximalanzahl der Circuits    */
        nmbcir_max = nmbcir;
    }
  } else {                             /* Verbindungsabbruch            */
    if (cirpoi->state != L4SDSCED) {   /* Circuit wird inaktiv          */
      ulink((LEHEAD *)cirpoi);         /* Circuit aus der Aktivliste    */
      relink((LEHEAD *)cirpoi,         /* in die Liste fuer unbenutzte  */
             (LEHEAD *)l4frel.tail);   /* Controllbloecke               */
      nmbcir--;                        /* einen aktiven Circuit weniger */
    }
  }
  cirpoi->state = state;               /* neuen Status setzen           */
}

/*----------------------------------------------------------------------*/
void l4tx(void)                        /* Frames senden                 */
{
  UWORD unack;                         /* unbestaetigte Frames          */
  UWORD isweg;                         /* schon gesendete Frames        */
  MBHEAD *fbp;                         /* naechstes Frame               */
  CIRBLK *nxtcir;                      /* Zeiger auf naechsten CIRBLK   */

  for (cirpoi = (CIRBLK *) l4actl.head;/* alle aktiven Circuits         */
       cirpoi!= (CIRBLK *)&l4actl.head;/* durchgehen                    */
       cirpoi = nxtcir)                /* und zum naechsten Circuit     */
  {
    nxtcir = (CIRBLK *) cirpoi->head;  /* wegen l4newstate()...         */

    if ((cirpoi->state == L4SIXFER) &&     /* Eintrag hat L4-Verbindung */
        (!(cirpoi->l4flag & L4FPBUSY))) {  /* und Partner nicht choked  */

      unack = (cirpoi->l4vs - cirpoi->l4rxvs) & 0x7F;
      if ((unack < cirpoi->numtx)          /* nur wenn Fenstergroesse   */
          && (unack < cirpoi->window)) {   /* noch nicht erreicht       */

        fbp = (MBHEAD *) cirpoi->mbhdtx.head; /* Anfang der Sendeliste  */
        for (isweg = 0; isweg < unack; ++isweg)/* schon gesendete offene*/
          fbp = (MBHEAD *) fbp->nextmh;        /* Frames, die auf Be-   */
                                               /* staetigung warten,    */
                                               /* uebergehen            */
        do {
          fbp->l4trie = 0;                    /* noch kein Versuch      */
          sndfrm(cirpoi->l4vs++, fbp);        /* naechstes Frame senden */

          fbp = (MBHEAD *) fbp->nextmh;       /* naechstes Frame        */
        } while ((++unack < cirpoi->numtx)    /* bis keine Frames mehr  */
                 &&(unack < cirpoi->window)); /* o.Fenstergroesse ueber-*/
                                              /* schritten              */
        if (!cirpoi->RTT)                     /* keine Messung unterwegs*/
          l4rtt(START);                       /* dann neue starten      */
      }
    }
  }
}

/*----------------------------------------------------------------------*/
/* Diese Routine prueft, ob das aktuelle Frame ein Circuit von uns sein */
/* koennte. Das ist ziemlich unsauber, eigentlich sollte nur im L4      */
/* landen, was auch direkt an uns addressiert ist. Das Lokal-Konzept    */
/* von alten NET/ROMs zwingt uns aber, auch "fremde" Frames zu durch-   */
/* suchen.                                                              */
BOOLEAN l4istome(char *srcid, char *dstid)
{
  if (l4opco != L4CONREQ)             /* es ist kein Connect Request    */
  {
    if (   (l4hdr0 < NUMCIR)          /* Index im richtigen Bereich?    */
                                      /* die Verbindung haben wir       */
        && (cirtab[l4hdr0].state != L4SDSCED)
                                      /* und die ID stimmt              */
        && (cirtab[l4hdr0].ideige == l4hdr1))
    {
      cirpoi = &cirtab[l4hdr0];       /* CIRCUIT-Eintrag ist gueltig!   */
      if (l4opco != L4CONACK)         /* CONACK kann vom Host kommen    */
          if (!cmpid(srcid, cirpoi->l3node->id))
              return (FALSE);         /* Absender muss stimmen          */
      if (!cmpid(dstid, myid))        /* nicht direkt an uns ?          */
          if (!cmpid(dstid, cirpoi->destca))
              return (FALSE);         /* Ziel muss stimmen              */
      return (TRUE);
    }
  }
  return (FALSE);
}

/*----------------------------------------------------------------------*/
void l4rx(NODE *srcnod, NODE *dstnod, MBHEAD *fbp)   /* Frame empfangen */
{
  char usrcal[7];                     /* Call des Users                 */
  char orgnod[7];                     /* Call des absender Knotens      */
  UWORD fenste;                       /* Fenstergroesse                 */
  int i;                              /* Sratch Zaehler                 */
  MBHEAD *antwor;                     /* Antwort auf das empfange Frame */
  CIRBLK *cirent;                     /* Eintrag des Users in der CIRTAB*/
  CIRBLK  nocir;                      /* Platzhalter wenn kein CIRCUIT  */
  char *viapoi;                       /* Ziger in der VIA-Liste         */
  char upno[L2IDLEN];                 /* Uplink Knoten                  */
  char upnov[L2VLEN+1];               /* L2-via zu dem Uplink Knoten    */

  /* hier gab es oefters Abstuerze, da itol3() immer einen gueltigen    */
  /* cirpoi erwartet, aber wir haben eventuell keinen, weil wir die     */
  /* Verbindung nicht kennen. Dies betrifft alle Versionen von TheNet   */

  (cirpoi = &nocir)->l3node = srcnod; /* Antwortweg wenn kein Circuit   */

  if (l4opco != L4CONREQ)             /* es ist kein Connect Request    */
  {
    if (   (l4hdr0 < NUMCIR)          /* Index im richtigen Bereich?    */
                                /* die gewuenscht Verbindung haben wir  */
        && (cirtab[l4hdr0].state != L4SDSCED)
                                      /* und die ID stimmt              */
        && (cirtab[l4hdr0].ideige == l4hdr1))
    {
      cirpoi = &cirtab[l4hdr0];       /* CIRCUIT-Eintrag ist gueltig!   */
      l4pidx = cirpoi->idxpar;        /* Partner-Index uebernehmen      */
      l4pcid = cirpoi->idpart;        /* Partner-ID uebernehmen         */
    }
    else
    {
      if (l4opco == L4CONACK)         /* alles ausser CONACK = Muell    */
      {
        l4pidx = l4hdr2;              /* Partner Circuit-Index          */
        l4pcid = l4hdr3;              /* und -ID fuer Antwort           */
        l4ahd2 =
        l4ahd3 = 0;
        l4aopc = L4DISREQ;            /* Disconnect Request als Antwort */
        itol3(1,gennhd());            /* Antwort &nocir->l3node=despoi  */
      }
      dealmb(fbp);                    /* schlechter Header, entsorgen   */
      return;
    }
  }

  switch (l4opco)                     /* Ueber Opcode verzweigen        */
  {
    case L4CONREQ:                    /* Connect-Request oder ein       */
      if (((fbp->mbpc - fbp->mbgc) >= 15) &&      /* Frame lang genug ? */
          ((fenste = getchr(fbp) & 0x7F) != 0) && /* und Fenster da     */
          (getfid(usrcal, fbp) == TRUE) &&  /* gueltiges Usercall       */
          (getfid(orgnod, fbp) == TRUE))    /* gueltiger Absender       */
      {
/*****************************************************************************\
*                                                                             *
* Protokollerweiterung: Beim Connect-Request-Frame wird im Anschluss an die   *
* Fenstergroesse der Uplinkknoten und die Via-Liste beim Uplink uebertragen   *
* (nullterminiert), also maximal 64 Bytes zusaetzlich. Diese Erweiterung ist  *
* kompatibel zur bisherigen Software, da laengere Frames nicht weiter unter-  *
* sucht werden. Um auch hier spaetere Erweiterungen zu ermoeglichen, wird     *
* der Rest des Frames nicht untersucht. Ist kein Uplinkknoten im Frame        *
* enthalten, also bei aelterer Software, wird das Call des Absenderknotens    *
* als Uplinkknoten eingesetzt.                                                *
*                                                                             *
\*****************************************************************************/
          if (   !(fbp->mbpc - fbp->mbgc >= L2IDLEN)  /* neue Software?       */
              || (!getfid(upno,fbp)))              /* kein Uplinkknoten?      */
          {
            cpyid(upno,orgnod);  /* Absenderknoten als Uplinkknoten annehmen  */
            *upnov = '\0';       /* und dort keine Digikette                  */
          }
          else                   /* neue Software                             */
          {
            for (viapoi = upnov;                 /* Vialiste holen            */
                 viapoi < upnov + L2VLEN;        /* max. 8 Calls              */
                 viapoi = viapoi + L2IDLEN)      /* viapoi -> naechstes Call  */
            {
              if (!getfid(viapoi, fbp)) break;   /* Ende Vialiste?            */
            }
            *viapoi = '\0';               /* Ende markieren                   */
          }

          l4pidx = l4hdr0;        /* Index                            */
          l4pcid = l4hdr1;        /* und ID des Parnters merken       */
          cirent = NULL;
          for (i = 0, cirpoi = cirtab; /* Circuit Tabelle absuchen    */
               i < NUMCIR;++i, ++cirpoi)
          {
            if (cirpoi->state != L4SDSCED) {  /* Verbingung besteht   */
              if (   (cirpoi->idxpar == l4hdr0) /* PartnerIndex stimmt*/
                  && (cirpoi->idpart == l4hdr1) /* Parner-ID stimmt   */
                  && cmpid(usrcal, cirpoi->upcall) /* UserCall stimmt */
                  && cmpid(orgnod, cirpoi->downca))/* Absender stimmt */
                 break;           /* wir haben ihn gefunden !         */
            }
            else                  /* sonst merken wir uns den freien  */
              if (cirent == NULL) /* Platz in der Circuit Table       */
                cirent = cirpoi;
          }
          if (i == NUMCIR) {      /* Eintrag war nicht in der Liste   */
            if ((cirent != NULL) && /* wenn wir noch ein Plaetzchen   */
                (fvalca(VCpar, usrcal) == TRUE)) /* haben und Call ok */
            {
              cirpoi = cirent;      /* dann Eintrag nehmen            */
              cpyid(cirpoi->upcall, usrcal);   /* Usercall setzen     */
              cpyid(cirpoi->downca, orgnod);   /* Absenderknoten      */
              cpyid(cirpoi->upnod,  upno);     /* Uplink Knoten       */
              cpyidl(cirpoi->upnodv, upnov); /* Uplink via's          */
              cirpoi->idxpar = l4hdr0;       /* Partner-Index merken  */
              cirpoi->idpart = l4hdr1;       /* Parner-ID merken      */
              cirpoi->ideige = rand() % 256; /* eigene Zufalls-ID     */
              cirpoi->l3node = srcnod;       /* Nachbar merken        */
              if (fbp->l3_typ == L3LOCAL)    /* ein Lokal?            */
                cpyid(cirpoi->destca, dstnod->id);
              else
                cpyid(cirpoi->destca, myid);

              cirpoi->tranoa = ininat;    /* Timeout setzen           */
            }
            else
            {                    /* kein Platz oder ungueltiges Call  */
              l4ahd2 =           /* Antwort aufbauen                  */
              l4ahd3 = 0;
              l4aopc = L4CONACK | L4CCHOKE;  /* Antwortframe aufbauen */
              (antwor = gennhd())->l2link = (LNKBLK *) srcnod;
              relink((LEHEAD *)antwor,(LEHEAD *) l3txl.tail);
              break;            /* und in die Sendekette haengen.     */
            }
          }                     /* den Eintrag gibt es schon          */
          /* Vom Partner vorgeschlagene Fenstergroesse uebernehmen    */
          /* wenn sie nicht groesser als unsere maximal-Groesse ist.  */
          cirpoi->window = (fenste > trawir) ? trawir : fenste;
          clrcir();               /* Eintrag initialisieren           */
          l4ahd2 = (UBYTE) (cirpoi - cirtab); /* Unseren Index setzen */
          l4ahd3 = cirpoi->ideige;            /* Unsere (Zufalls-) ID */
          l4aopc = L4CONACK;                  /* Connect Acknowledge  */
          putchr(cirpoi->window, (antwor = gennhd())); /* endgueltige */
                             /* Fenstergroesse zurueck an den Partner */
          itol3(1,antwor);                    /* senden               */
          switch (cirpoi->state) {
            case L2SDSCED :
            case L2SLKSUP :
              l4newstate(L4SIXFER);           /* STATUS = connected   */
              l2tol7(L4MCONNT, cirpoi, L4_USER);
          }
      }
      break;

      case L4CONACK:
        if (cirpoi->state == L4SLKSUP) {   /* nur wenn Connect von uns  */
          if ( !(l4hdr4 & L4CCHOKE) ) {    /* verlangt war und Partner  */
            if (fbp->mbgc < fbp->mbpc) {   /* nicht choked.             */
              cirpoi->window = getchr(fbp); /* Entgueltige Fenstergr.   */
              cirpoi->idpart = l4hdr3;     /* Partner ID                */
              cirpoi->idxpar = l4hdr2;     /* Partner Index             */
              clrcir();                    /* Eintrag initialisieren    */
              cirpoi->tranoa = ininat;     /* Timeout setzen            */
              l4newstate(L4SIXFER);        /* Status = connected        */
              /**********************************************************/
              /* Protokollerweiterung nach DB7KG, 23.11.1996            */
              /* Bei jedem eingehenden Connect-ACK setzen wir den Ziel- */
              /* Node auf den Absender, von dem das Frame kam. Damit    */
              /* soll bei einem Connect an einen Local der tatschliche */
              /* L3 Partner als Ziel gewaehlt werden, damit es keine    */
              /* Probleme beim umrouten dieser Verbindungen gibt.       */
              /* (Ein umrouten ist dann nur bis zum Partner moeglich,   */
              /* da kuenftig Frames mit dieser Addresse gesendet werden)*/
              /* HINWEIS: Impelementierung hier so nur moeglich, weil   */
              /* l4rx() direkt aus dem L3 aufgerufen wird! (fr DF2AU)  */
              /**********************************************************/
              cirpoi->l3node = srcnod;     /* in Zukunft an den Node    */
              l2tol7(L4MCONNT, cirpoi, L4_USER); /* L7 melden           */
            }
          }
          else                             /* Partner ist choked (busy) */
              l4nsta(L4MBUSYF);            /* an L7 melden              */
        }
        break;

      case L4DISREQ:                       /* Disconnect Request        */
        l4ahd2 =
        l4ahd3 = 0;                        /* Antwort aufbauen          */
        l4aopc = L4DISACK;                 /* Opcode: Disconnect Ackn.  */
        itol3(1,gennhd());                 /* Frame senden              */
        clr4rx(1);                         /* restliche Infos senden    */
        l4nsta(L4MDISCF);                  /* an L7 melden              */
        break;

      case L4DISACK:                       /* Disconnect Acknowledge    */
        if (cirpoi->state == L4SDSCRQ)     /* Hatten wir Disc gesendet ?*/
          l4nsta(L4MDISCF);                /* wenn ja dann dem L7 melden*/
        break;

      case L4INFTRA:                       /* Infoframe                 */
        if (cirpoi->state != L4SIXFER)     /* nur wenn connected        */
          break;
        chksts();                          /* Status Info auswerten     */
        if (((fbp->l2fflg =                /* passt Frame ins Fenster   */
          (l4hdr2 - cirpoi->l4vr) & 0x7f) < cirpoi->window)
          && !(cirpoi->l4flag & L4FBUSY))
        {                                  /* und Partner nicht busy    */
          fbp->morflg = (l4hdr4 & L4CMORE) != 0;  /* Fragmentiert ?     */
          if (fbp->l2fflg == 0) {          /* passt Sequenz?            */
            takfrm(fbp);                   /* Frame uebernehmen         */
            /* Frames, die ausser der Reihe kamen ueberpruefen          */
            for (i = 1, antwor = (MBHEAD *) cirpoi->mbhdos.head;
                 (MBHEAD *) &(cirpoi->mbhdos) != antwor;
                 antwor = (MBHEAD *) antwor->nextmh)
            {
              if ((antwor->l2fflg -= i) == 0) {  /* passt das Frame ?   */
                  fbp = (MBHEAD *) antwor->prevmh; /* ja, nehmen und    */
                  takfrm((MBHEAD *) ulink((LEHEAD *)antwor));/*naechstes*/
                  antwor = fbp;     /* Frame ueberpruefen               */
                  ++i;              /* Offset eins weiter               */
              }
            }
            cirpoi->l4rs = 0;      /* Antwort: ACK (Bestaetigung)       */

            /* Dynamisches ACKDEL */

            i =  cirpoi->numrx - (cirpoi->window /2);
            if ( i < 1 ) i = 1;
            cirpoi->acktim = i * L4_ACKDEL ;
          }
          else
          {                        /* Sequenz passt nicht              */
            for (antwor = (MBHEAD *) cirpoi->mbhdos.head;;) {
              /* Wenn die Kette dort zuende ist, dann dort einhaengen */
              if ((MBHEAD *) &(cirpoi->mbhdos) == antwor) {
                relink((LEHEAD *)fbp, (LEHEAD *)cirpoi->mbhdos.tail);
                break;
              }
              /* Wenn das Frame schoneinmal gekommen ist -> weg damit */
              if (antwor->l2fflg == fbp->l2fflg) {
                dealmb(fbp);
                break;
              }
              /* sonst an der passenden Stelle einhaengen             */
              if (antwor->l2fflg > fbp->l2fflg) {
                relink((LEHEAD *)fbp, (LEHEAD *)antwor->prevmh);
                break;
              }
              antwor = (MBHEAD *) antwor->nextmh; /* ein Frame weiter */
            }

            if (cirpoi->l4rs == 0) {      /* Wenn ACK gesendet werden soll */
              cirpoi->l4rs = 1;           /* dann nun einen N(ot)AK senden */
              cirpoi->acktim = L4_ACKDEL; /* ACK-Wartezeit setzen          */
            }
          }
          return;                     /* Frame verarbeitet                */
        }
        else                          /* ungueltiges Frame oder wir haben */
          cirpoi->acktim = L4_ACKDEL; /* zuwenig Buffers (choked), dann   */
        break;                        /* Antwort verzoegern.              */

      case L4INFACK:                  /* Info Acknowledge                 */
        if (cirpoi->state == L4SIXFER)/* nur wenn connected               */
          chksts();                   /* Statusinformation auswerten      */
   }
   dealmb(fbp);                       /* hier landen wir bei ungueltigen  */
                                      /* Opcodes oder wenn die Bearbeitung*/

                                      /* fertig ist, auf jeden Fall Frame */
}                                      /* wegwerfen.                      */

/*------------------------------------------------------------------------*/
void l4rest(void)                     /* sonstige L4-Funktionen           */
{
  UWORD   rx_unack;
  UBYTE   w2;
  CIRBLK *nxtcir;

  for (cirpoi =(CIRBLK *) l4actl.head; /* alle aktiven Circuits         */
       cirpoi!=(CIRBLK *)&l4actl.head; /* durchgehen                    */
       cirpoi = nxtcir) {
    nxtcir = (CIRBLK *) cirpoi->head;  /* vorher schon merken           */
    if (   (cirpoi->l4flag & L4FDIMM)) /* sofortiger Abwurf             */
    {
         l4nsta(L4MDISCF);             /* Disconnect melden             */
    }
    else
    if (cirpoi->state == L4SIXFER) {/* nur fuer connectete Eintraege    */
      if (   (cirpoi->l4flag & L4FDSLE) /* Abwurf gefordert             */
          && (cirpoi->numtx == 0) )   /* und alle Infos gesendet        */
        endcir();                    /* Eintrag loeschen                */
      else
      {
        clr4rx(0);                   /* sonst Info senden               */

        /* Fruehes Ack */
        rx_unack = ( (cirpoi->l4vr | 0x100) - cirpoi->ll4txNR ) & 0x7F;
        w2 = cirpoi->window/2;

        if ( !(cirpoi->l4flag & L4FBUSY) ) { /* wir sind nicht busy     */
          if ( nmbfre < 30) {                /* Aber zu wenig Platz     */
            cirpoi->l4flag |= L4FBUSY; /* Dann sind wir nun Busy !      */
            cirpoi->l4rs = 0;          /* Antwort sofort senden         */
            sndack();                  /* ACK senden                    */
          }
          else {                       /* ich bin nicht busy            */
            if (   cirpoi->acktim > 1
                && rx_unack >= w2
                && cirpoi->numrx < w2    /* und es ist noch Platz       */
               )
              sndack();
          }
        }
        else {  /* im Moment sind wir Busy */
          if (    nmbfre > 62             /* wieder genug Platz ?      */
              &&  cirpoi->numrx < w2       /* Und nicht zuviel ?        */
          ) {
            cirpoi->l4flag &= ~L4FBUSY;    /* Busy aufheben              */
            sndack();                      /* und ACK senden             */
          }
        }
      }
    }
  }
}

/*----------------------------------------------------------------------*/
void trasrv(void)                   /* Timerservice fuer den L4         */
{
  UWORD   actsts;                   /* Status des aktuellen Eintrages   */
  UWORD   fropen;                   /* Zahl der unbestaetigten Frames   */
  UWORD   tosend;                   /* Zahl der zu sendenden Frames     */
  MBHEAD *fbp;                      /* aktuelles Frame                  */
  UWORD   rx_unack;
  CIRBLK *nxtcir;
  char    txvs;

  for (cirpoi =(CIRBLK *) l4actl.head; /* alle aktiven Circuits         */
       cirpoi!=(CIRBLK *)&l4actl.head; /* durchgehen                    */
       cirpoi = nxtcir) {
    nxtcir = (CIRBLK *) cirpoi->head;  /* vorher schon merken           */
    if ((actsts = cirpoi->state) != L4SDSCED) { /* nur fuer aktive Verb.*/
      if (cirpoi->RTT)              /* RTT-Messung                      */
        cirpoi->RTT++;
      if (cirpoi->traout != 0) {    /* Timeout noch nicht abgelaufen?   */
        if (--cirpoi->traout == 0) {/* Timeout nun abgelaufen ?         */
          if (actsts == L4SIXFER) { /* nur fuer connectete Eintraege    */
            cirpoi->l4flag &= ~L4FPBUSY; /* nichts mehr senden          */
          } else {
            if (++cirpoi->l4try < L4_RETRY) /* nochmal Versuchen ?      */
            {
              if (actsts == L4SLKSUP)  /* CON REQ kam nicht an, nochmal */
                sconrq();
              else
                sdisrq();             /* DISC REQ kam nicht an, nochmal */
            } else
              l4nsta(L4MFAILW);       /* Fehler an L7 melden            */
          }
        }
      }
      else {                          /* Timeout ist abgelaufen         */
        if ((actsts == L4SIXFER) /* connected und Frames unbestaetigt ? */
             && ((fropen = (cirpoi->l4vs - cirpoi->l4rxvs) & 0x7F) != 0)) {
                                             /* Frames wiederholen      */
          for (tosend = 0, fbp = (MBHEAD *) cirpoi->mbhdtx.head;
               tosend < fropen;
             ++tosend, fbp = (MBHEAD *) fbp->nextmh) {
            if (--fbp->l4time == 0) {        /* wenn Timeout um         */
              if (++fbp->l4trie < L4_RETRY) {/* und noch Versuch frei   */
                txvs = fbp->l2fflg;          /* Framenummer holen       */
                if (cirpoi->RTTvs == txvs)   /* bei Wiederholung RTT-   */
                  l4rtt(CLEAR);              /* Messung verwerfen       */
                sndfrm(txvs, fbp);           /* Frames senden           */
              } else {
                l4nsta(L4MFAILW);            /* sonst L7 failure melden */
                break;
              }
            }
          }
        }
      }

      if (actsts == L4SIXFER) {        /* connected ?                    */
        if (   (cirpoi->acktim != 0)   /* ACK noch nicht gesendet        */
            && (--cirpoi->acktim == 0) /* aber Timer laeuft gerade aus   */
            )
        {
          /* schnelle Flow-control */
          rx_unack = ( (cirpoi->l4vr | 0x100) - cirpoi->ll4txNR ) & 0x7F;

          if (   cirpoi->numrx > (cirpoi->window /2)
              && rx_unack >= cirpoi->window
             )
          {
            cirpoi->l4flag |= L4FBUSY;
            cirpoi->l4rs = 0;
          }
          sndack();                 /* dann ein ACK senden              */
        }

        if (   (cirpoi->tranoa != 0) /* No-activity Timeout laeuft aus   */
             && (--cirpoi->tranoa == 0)
           )
             endcir();               /* Verbindung trennen               */
      }
    }
  }
}

/*----------------------------------------------------------------------*/
void newcir(void)                   /* neuen Circuit-Eintrag aufbauen   */
{
  clrcir();                         /* Eintrag erstmal loeschen         */
  cirpoi->ideige = rand() % 256;    /* eigene ID zufaellig erzeugen     */
  cirpoi->l4try = 0;                /* noch keine Sende-Versuche gemacht*/
  sconrq();                         /* Connect Request senden           */
  l4newstate(L4SLKSUP);             /* neuer Status: Verbindungsaufbau  */
}

/*----------------------------------------------------------------------*/
void discir(void)                   /* Circuit aufloesen                */
{
  if ((cirpoi->state == L4SLKSUP) ||/* Status Connect Request oder      */
        (cirpoi->state == L4SDSCRQ))  /* Disconnect Request ?           */
  {
     /* Bis jetzt wurde ein L4 Disconnect nicht nach oben gemeldet,     */
     /* wenn der State noch LINK SETUP oder DISCONNECT REQUEST war.     */
     /* Ich nehme an, das war wegen der reentranz-Problematik so        */
     /* geloest worden. Jetzt wird ein Flag gesetzt, das eigenliche     */
     /* Melden passiert in L4rest.                                      */
     cirpoi->l4flag |= L4FDIMM;      /* spaeter in l4rest               */
  }
  else {
     kilfra();                       /* Fragmentliste loeschen          */
     dealml((LEHEAD *)&cirpoi->mbhdrx); /* Empfangsliste loeschen       */
     cirpoi->numrx = 0;              /* Empfangszaehler zuruecksetzen   */
     cirpoi->l4flag |= L4FDSLE;      /* fuer Abwurf markieren           */
  }
}

/*----------------------------------------------------------------------*/
BOOLEAN itocir(BOOLEAN cflg, MBHEAD *mbp) /* Info an Circuit senden     */
{
  CIRBLK *cblk;

  if (((cblk = (CIRBLK *) mbp->l2link)->numtx < conctl) || /* Platz ?   */
        (cflg == TRUE)) {                            /* oder immer senden */
     /* Infoframe in die TX-Liste des Circuit haengen */
     relink(ulink((LEHEAD *)mbp),(LEHEAD *) cblk->mbhdtx.tail);
     ++cblk->numtx;                  /* Framezaehler erhoehen            */
/* Das morflg des Frames wird jetzt im L7 geloescht vor dem Aufruf von  */
/* itocir, falls das Frame nicht von einem L4-Partner kommt.            */
     cblk->tranoa = ininat;          /* Timeout neu aufziehen            */
     return(TRUE);                   /* ok zurueckmelden                 */
  }
  return(FALSE);                    /* kein Platz zurueckmelden         */
}

void clr4rx(conctl)    /* Informationstransfer von Layer 4 nach Layer X */
/* Solange noch empfangene Pakete vorhanden sind, werden diese an       */
/* andere Layer durch Aufruf von fmlink() uebertragen. Bei geseztem     */
/* Ueberfuellungskontroll-Flag (conctl == TRUE) wird die Uebertragung   */
/* abgebrochen, wenn der andere Layer keine weiteren Daten mehr auf-    */
/* aufnehmen kann.                                                      */
/* Nach erfolgter Uebertragung wird die Anzahl der uebertragenen Zeichen*/
/* fuer die Statistik gezaehlt und der No-Activity-Timer neu gesetzt.   */
BOOLEAN conctl;
{
    MBHEAD *mbp;

    while (cirpoi->numrx != 0) {/* solange noch Frames vorhanden    */
      mbp = (MBHEAD *) cirpoi->mbhdrx.head; /* ein Frame holen      */
      mbp->l2link = (LNKBLK *) cirpoi; /*zugehoerigen Circuit merken*/
      mbp->type = 4;                   /* User ist Circuit          */
      mbp->l2fflg = L2CPID;            /* L4 immer PID F0           */
      if (!fmlink(conctl, mbp)) break; /* Ende bei Fehler           */
      --cirpoi->numrx;                 /* ein Frame weniger         */
      cirpoi->tranoa = ininat;         /* Timeout neu setzen        */
      l4setT2();                       /* ACK_TIMER neu setzen      */
    }
}

/*----------------------------------------------------------------------*/
void chksts(void)                   /* Status des Frames auswerten      */
{
  UWORD frofs;                      /* bestaetigter Offset              */
  UWORD fropen;                     /* unbesteatigter Offset            */

  if ((fropen = (cirpoi->l4vs - cirpoi->l4rxvs) & 0x7F) != 0) {
                                    /* Frames offen ?                   */
    if ((frofs = (l4hdr3 - cirpoi->l4rxvs) & 0x7F) != 0) {
                                    /* neu bestaetigte Frames ?         */
      if (frofs <= fropen) {        /* Frames wurden bestaetigt         */
        while (frofs-- != 0) {      /* solange bestaetigte Frames ueber */
          dealmb((MBHEAD *)ulink((LEHEAD *)cirpoi->mbhdtx.head)); /* weg*/
          cirpoi->numtx--;          /* einen weniger zu senden          */
          if (cirpoi->l4rxvs++ ==   /* einen mehr bestaetigt            */
              cirpoi->RTTvs)        /* Messframe bestaetigt?            */
            l4rtt(STOP);            /* Messung auswerten                */
        }
      }
    }
  }

  if ( !(l4hdr4 & L4CCHOKE) ) {      /* Partner choked ?                 */
     cirpoi->l4flag &= ~L4FPBUSY;    /* nein, merken                     */
     l4clrT3();                      /* Timeout kann nicht kommen        */
     if ((l4hdr4 & L4CNAK) &&        /* NAK-Flag ?                       */
          (cirpoi->l4vs != cirpoi->l4rxvs)) { /* und noch was offen ?    */
                                              /* dann wiederholen        */
       if (cirpoi->RTTvs == cirpoi->l4rxvs)
         l4rtt(CLEAR);               /* bei Wiederholung RTT verwerfen   */
       sndfrm(cirpoi->l4rxvs, (MBHEAD *)cirpoi->mbhdtx.head);
     }
  }
  else { /* Partner ist choked (busy) */
     cirpoi->l4flag |= L4FPBUSY;     /* merken                           */
     cirpoi->l4vs = cirpoi->l4rxvs;  /* keine Frames offen               */
     l4setT3(L4TBUSY);               /* warten                           */
  }
}

/*----------------------------------------------------------------------*/
void takfrm(MBHEAD *mbp)            /* empfangenes Frame uebernehmen    */
{
  BOOLEAN more;         /* morflg des zu verarbeitenden Frames          */
  BOOLEAN copy = FALSE; /* umkopieren noetig wegen Fragmentierung       */
  MBHEAD *fragmn;       /* dieses Fragment wird bearbeitet              */
  int     max;          /* maximale Info-Laenge fuer Weiterleitung      */
  int     sum  = 0;     /* Summe vorhandenes Fragment + neues Frame     */

  if ( !(cirpoi->l4flag & L4FDSLE)) /* Circuit ist nicht zu beenden     */
   {
    more = mbp->morflg;

/* Framegroesse fuer vorhandenes Fragment und neues Frame bestimmen.    */
/* Da fuer das Fragment in einen neuen Buffer geschrieben wurde, ist    */
/* (fragmn->mbgc == 0) und muss nicht von fragmn->mbpc subtrahiert      */
/* werden.                                                              */

    if ((fragmn = cirpoi->fragme) != NULL)
     {
      sum += fragmn->mbpc;
     }
    sum += mbp->mbpc - mbp->mbgc;
     
/* Im L7 nachfragen, welche maximale Paketlaenge verarbeitet werden     */
/* kann (CCP / Hostmode / L2 -> 256 Bytes; L4 -> 236 Bytes)             */

    max = ptc_p_max(cirpoi, L4_USER);

/* muss umkopiert werden?                                               */
    if ((sum > max) || (fragmn != NULL) || (more))
      copy = TRUE;

    while (copy)
     {
/* ggf. neuen Buffer fuer Arbeits-Fragment holen                        */
      if (fragmn == NULL)
       {
        fragmn = (MBHEAD *) allocb();
        cirpoi->fragme = fragmn;
       }
/* erstmal vermuten, dass nach diesem Fragment noch mehr folgt          */
      fragmn->morflg = 1;
/* aus neuem Frame ans Fragment anhaengen, bis max erreicht ist, oder   */
/* neues Frame leer                                                     */
      while ((fragmn->mbpc < max) && (mbp->mbpc > mbp->mbgc))
       {
        putchr(getchr(mbp), fragmn);
       }
/* Fragment fertig - Rest berechnen                                     */
      sum -= fragmn->mbpc;
      if (!sum)
       {
/* neues Frame ist leer und kann weg                                    */
        dealmb(mbp);
        mbp = NULL;
        copy = FALSE;
/* wenn nichts mehr folgt, dies im Fragment vermerken                   */
        if (!more)
          fragmn->morflg = 0;
       }
/* Fragment weiterreichen, wenn max erreicht oder wenn nix mehr kommt   */
      if ((fragmn->mbpc == max) || (fragmn->morflg == 0))
       {
        rwndmb(fragmn);
        relink((LEHEAD *)fragmn, (LEHEAD *) cirpoi->mbhdrx.tail);
        fragmn = NULL;
        cirpoi->fragme = NULL;
        ++cirpoi->numrx;
       }
     }

    if (mbp != NULL)               /* Frame noch da?                   */
     {
      relink((LEHEAD *)mbp,(LEHEAD *) cirpoi->mbhdrx.tail); /* in L4-RX */
      ++cirpoi->numrx;              /* ein Frame mehr empfangen         */
     }
   }
  else                              /* Verbindung soll abgeworfen werden*/
    dealmb(mbp);                    /* Frame einfach vernichten         */
  ++cirpoi->l4vr;                   /* RX-Sequenz erhoehen              */
}

/*----------------------------------------------------------------------*/
void sndfrm(int txsequ, MBHEAD *mbp)   /* Frame senden*/
{
  char huge *next;                  /* Pointer auf Ende des Headers     */
  MBHEAD *netmhd;                   /* Netzwerk Header                  */
  int mtu;                          /* MTU auf dem Sende-Port           */

  l4ahd2 =                          /* Sendesequenzzaehler in den Header*/
  mbp->l2fflg = txsequ;             /* und im Buffer merken             */
  l4aopc = L4INFTRA;                /* Opcode: Info                     */
  if (mbp->morflg)
    l4aopc |= L4CMORE;
  ackhdr();                         /* Rest des Headers erzeugen        */
  next = (netmhd = gennhd()) ->mbbp;/* Buffer holen und Position Opcode */
                                    /* merken                           */
  mtu = 256;                        /* NETROM immer 256 Bytes Frames    */

  if (splcpy((mtu - netmhd->mbpc), netmhd, mbp)) /* Info umkopieren     */
  {                                 /* hat nicht alles reingepasst      */
    ++cirpoi->numtx;                /* ein Frame mehr gesendet          */
    mbp->morflg = TRUE;             /* markieren, es kommt noch mehr    */
  }

  if (mbp->morflg)
     *(next -1) |= L4CMORE;         /* more Flag im Opcode setzen       */

  itol3(0,netmhd);                  /* Frame an L3 geben                */
  l4setT1(mbp);                     /* Timeout neu aufziehen            */
}

/*----------------------------------------------------------------------*/
void l4nsta(WORD frtyp)             /* Statusaenderung im L4            */
{
  l4newstate(L4SDSCED);             /* Status: Disconnected             */
  clrcir();                         /* Eintrag zuruecksetzen            */
  dealml((LEHEAD *) &cirpoi->mbhdrx);  /* Empfangsliste loeschen        */
  dealml((LEHEAD *) &cirpoi->mbhdtx);  /* Sendeliste loeschen           */
  cirpoi->numrx =                   /* Zaehler zuruecksetzen            */
  cirpoi->numtx = 0;
  l2tol7(frtyp, cirpoi, L4_USER);
}

/*----------------------------------------------------------------------*/
void endcir(void)               /* Circuit aufloesen                    */
{
  clrcir();                     /* Eintrag in CIRTAB loeschen           */
  cirpoi->l4try = 0;            /* Versuche zuruecksetzen               */
  sdisrq();                     /* Abwurf einleiten                     */
  l4newstate(L4SDSCRQ);             /* neuer Status: DISC-REQ gegeben       */
}

/*----------------------------------------------------------------------*/

void clrcir(void)               /* Eintrag in CIRTAB loeschen           */
{
  kilfra();                     /* Fragmente loeschen                   */
  dealml((LEHEAD *)&cirpoi->mbhdos);    /* Messageliste dafuer auch     */
  cirpoi->l4rxvs =              /* alle Sequenzen auf 0                 */
  cirpoi->l4vs =
  cirpoi->l4vr =
  cirpoi->l4rs =                /* ACK-NAK Flag                         */
  cirpoi->l4try =
  cirpoi->ll4txNR  = 0;
  cirpoi->l4flag = 0;           /* niemand choked, kein DISC-REQ        */
  cirpoi->SRTT = L4_IRTT;       /* SRTT initialisieren                  */
  l4clrT3();                    /* Timer loeschen                       */
  l4rtt(CLEAR);                 /* RTT-Messung stoppen und verwerfen    */
  resptc(g_uid(cirpoi, L4_USER));
}

/*----------------------------------------------------------------------*/

void sconrq(void)               /* CON-REQ senden                       */
{
  MBHEAD *mbp;                  /* Buffer fuer Frame                    */
  char   *viapoi;               /* Zeiger in Vialiste                   */

  l4pidx = cirpoi - &cirtab[0]; /* Index setzen                         */
  l4pcid = cirpoi->ideige;      /* Partner und eigener Index            */
  l4ahd2 =                      /* zwei Bytes leer                      */
  l4ahd3 = 0;
  l4aopc = L4CONREQ;            /* Opcode                               */
  putchr (trawir, (mbp = gennhd()));    /* Rest des Headers             */
  putfid(cirpoi->upcall, mbp);  /* beide Calls in das Frame             */
  putfid(myid, mbp);
  putfid(cirpoi->upnod, mbp);   /* Uplinkknoten dazu                    */
  viapoi = cirpoi->upnodv;      /* Uplink-Vialiste auch                 */
  while (*viapoi != '\0')       /* Ende Vialiste erreicht?              */
    {
     putfid(viapoi, mbp);        /* Call in Puffer                       */
     viapoi = viapoi + L2IDLEN;  /* zum naechsten Call                   */
    }
  putchr('\0', mbp);            /* 0 markiert Ende Vialiste             */
  itol3(1,mbp);                 /* an Layer 3 liefern                   */
  l4setT3(L4TCREQ);             /* Timeout setzen                       */
}

/*----------------------------------------------------------------------*/

void sdisrq(void)               /* DISQ-REQ senden                      */
{
  l4pidx = cirpoi->idxpar;      /* Index setzen                         */
  l4pcid = cirpoi->idpart;      /* und ID                               */
  l4ahd2 =                      /* 2 Bytes leer                         */
  l4ahd3 = 0;
  l4aopc = L4DISREQ;            /* Opcode                               */
  itol3(1,gennhd());            /* Rest des Headers und dann an Layer 3 */
  l4setT3(L4TDREQ);             /* Timeout setzen                       */
}

/*----------------------------------------------------------------------*/
/* Fast L4 von DB7KG                                                    */
/*----------------------------------------------------------------------*/
void sndack(void)               /* ACK senden                           */
{
  l4ahd2 = 0x00;
  l4aopc = L4INFACK;            /* Opcode                               */
  ackhdr();                     /* Rest des Headers                     */
  itol3(2, gennhd());           /* schnelle Weiterleitung               */
}

/*----------------------------------------------------------------------*/

void ackhdr(void)               /* ACK Header erzeugen                  */
{
  l4pidx = cirpoi->idxpar;                      /* Partner Index        */
  l4pcid = cirpoi->idpart;                      /* Partner ID           */
  l4ahd3 = cirpoi->l4vr;                        /* RX-Sequenz           */

  cirpoi->ll4txNR  = cirpoi->l4vr & 0xFF;       /* Letzten ACK senden   */

  if (cirpoi->l4flag & L4FBUSY)                 /* selbst choked        */
     l4aopc |= L4CCHOKE;                        /* dann Flag setzen     */
  else {
     if (cirpoi->l4rs == 1) {           /* wird es ein NAK Header?      */
        l4aopc |= L4CNAK;               /* dann Flag setzen             */
        cirpoi->l4rs = 2;               /* NAK als gesendet markieren   */
     }
  }
  cirpoi->acktim = 0;                   /* ACK Timer ruecksetzen        */
}

/*----------------------------------------------------------------------*/

void itol3(WORD priority, MBHEAD *mbp)          /* Info an Layer 3      */
{
  mbp->l2link = (LNKBLK *) cirpoi->l3node;      /* Controllblock        */
  mbp->l4type = priority;                       /* Prioritt setzen     */
  relink((LEHEAD *)mbp,(LEHEAD *) l3txl.tail);  /* nur umhaengen        */
}

/*----------------------------------------------------------------------*/

MBHEAD *gennhd(void)            /* Netzwerk Header erzeugen             */
{
  int i;
  MBHEAD *mbp;                  /* Buffer fuer Info                     */

  mbp = (MBHEAD *) allocb();    /* Buffer besorgen                      */
  for (i = 0; i < 15; ++i)      /* die ersten 15 Bytes fuer Header leer */
     putchr(0, mbp);
  putchr(l4pidx, mbp);          /* Transport Header schreiben           */
  putchr(l4pcid, mbp);
  putchr(l4ahd2, mbp);
  putchr(l4ahd3, mbp);
  putchr(l4aopc, mbp);

  return(mbp);                  /* Buffer wird zurueckgegeben           */
}

/*----------------------------------------------------------------------*/

void kilfra(void)               /* Fragmente loeschen                   */
{
  if (cirpoi->fragme != NULL) {                 /* schon leer?          */
     dealmb(cirpoi->fragme);                    /* Fragment loeschen    */
     cirpoi->fragme = NULL;                     /* Eintrag loeschen     */
  }
}

/*-------------------------------------------------------------------------*/

void l3tol4(NODE *totnod)     /* Meldung L3 -> L4: Knoten, auf den despoi  */
{                             /* zeigt, wird aus der Nodesliste gestrichen */
  CIRBLK  *nxtcir;

  for (cirpoi =(CIRBLK *) l4actl.head; /* alle aktiven Circuits            */
       cirpoi!=(CIRBLK *)&l4actl.head; /* durchgehen, Eintraege suchen,    */
       cirpoi = nxtcir) {              /* die zum Zielknoten despoi gehen  */
    nxtcir = (CIRBLK *) cirpoi->head;  /* vorher schon merken              */
    if ( cirpoi->l3node == totnod ) {  /* Zielknoten wurde geloescht       */
      clr4rx(TRUE);      /* empfangene Frames abliefern                    */
      l4nsta(L4MFAILW);  /* Meldung an L7: Failure (Zielknoten ist weg)    */
    }
  }
}

/* End of L4.C */
